from PIL import Image, ImageFont, ImageDraw
import matplotlib.pyplot as plt
import numpy as np
import sys
import copy

colors_default = {
    'cyan': (0, 237, 255), 
    'magenta': (255, 0, 171), 
    'yellow': (255, 237, 0), 
    'red': (255, 0, 0),
    'blue': (0, 71, 171), 
    'green': (0, 181, 0),
    'perano':(167,184,248), # blue with slight purple
    'onahau':(204,238,255), # pale cyan
    'seagull':(136,204,238), # gray cyan
    'deco':(204,221,170), # gray yellow green
    'primerose':(238,238,187), # gray yellow
    'pink':(255,190,204), # pink
    'silver':(190,190,190), # light gray
    'apricot':(238,136,102), # orange
    'teal':(0, 153,136), # dirty cyan
    'cerise':(238,61,119), # violet red
    'purple':(170, 68,153), # Mulberry
    'olive':(170,170,0), # dirty yellow
    'wine':(136,34,85), # dirty pink
    'sundance':(201,181,81), # dirty pear
    'white':(255,255,255),
    'black':(0,0,0),
    'grey':(79,79,79)
}

abbr_dict = {
    'c':'cyan',
    'm':'magenta',
    'y': 'yellow', 
    'r':'red',
    'b':'blue',
    'g':'green',
    'w':'white',
    'd':'black'
}

for a in abbr_dict.keys():
    colors_default[a] = colors_default[abbr_dict[a]]


def CombineColor(pages, pagecolors = {}, bg = 'black', mode = 'layer', palette = {}, addlegend = True, normalize = True, colors_default = colors_default):
    '''
    Combine grayscale channels into one multicolor visualization.
    Returns a PIL image, which can be converted to np array by np.array(image).
    Use colorbar() to see all the colors.
    
    Inputs:
        pages: either a dict {pagename : pagedata}, or a 3d np array stacked along the 3rd axis.
        pagecolors: either a dict {pagename : colorname}, or a list/array of colornames. Channels not assigned colors will be colored automatically.
        bg: background color. Shown when all channels are 0. 
        mode: can be "layer", "blend" or "add".
            layer: overlapped areas shown as the latest channel.
            blend: overlapped areas shown as intermediate colors, as the colors are mixed. 
            add: overlapped areas shown as the added colors, as the colored beams add up as white light.
        palette: a dict {colorname : (r, g, b)}. rgb values are (0, 255). 
        addlegend: bool. whether adding legend at the bottom of the image. default = True.
        normalize: bool. If true, each channel will be independently max-normalized to (0,1) first. default = True.
        colors_default: Default palette. Set it to {} if only using custome colors.
    '''
    pages = copy.deepcopy(pages)
    pagecolors = copy.deepcopy(pagecolors)
    palette = copy.deepcopy(palette)
    colors_default = copy.deepcopy(colors_default)
    # pages_dict {pagename : pagedata}
    if hasattr(pages, 'keys'):
        pages_dict = pages
    elif hasattr(pages, 'shape'):
        pages_dict = {f'{i}': pages[:,:,i] for i in range(pages.shape[2])}
    else:
        raise TypeError("pages must be a dict or a 3D NP array.")
    pagenames = list(pages_dict.keys())
    
    
    # pagecolors {pagename : colorname}
    if not hasattr(pagecolors, 'keys'):
        pagecolors = { pagenames[i]:pagecolors[i] for i in range(len(pagecolors))}
    

    # if pagecolors is given in rgb instead of color names
    for k in pagecolors.keys():
        if not type(pagecolors[k]) == str:
            palette[f'{k}_color_'] = copy.deepcopy(pagecolors[k])
            pagecolors[k] = f'{k}_color_'

    # color_dict {colorname : rgb}
    color_dict = palette
    for c in colors_default.keys():
        if c not in color_dict.keys():
            color_dict[c] =  colors_default[c]
    
    colors = [c for c in list(color_dict.keys()) if c not in list(pagecolors.values()) + ['white','black','w','d']]

    # assign colors to undeclared pages.
    for pagename in pagenames:
        if pagename not in pagecolors.keys():
            pagecolors[pagename] = colors[0]
            colors = colors[1:]

    # pcolor_dict { pagename : rgb }
    for p in pagenames:
        if pagecolors[p] not in color_dict.keys():
            raise KeyError(f'{pagecolors[p]} RGB value not defined.')

    pcolor_dict = { p : color_dict[pagecolors[p]] for p in pagenames}

    # initialize
    im = np.zeros([pages_dict[pagename].shape[0], pages_dict[pagename].shape[1], 3])
    # layover
    
    if mode == 'blend':
        # devided by number of mixed colors at each pixel
        nn = np.dstack(list(pages_dict.values()))
        nn = np.sum(nn>0, axis=2)
        nn = np.where(nn == 0, 1, nn)

    for p in pagenames: # pagename
        D = pages_dict[p] # data
        D = D/max(1, np.max(D)) if normalize else D
        c = pcolor_dict[p] # rgb
        
        if mode =='layer':
            for j in [0,1,2]:
                im[:,:,j] = np.where(D>0, D*c[j], im[:,:,j])
        elif mode == 'blend':
            for j in [0,1,2]:
                im[:,:,j] = im[:,:,j] + D*c[j]/nn
        elif mode == 'add':
            for j in [0,1,2]:
                im[:,:,j] = im[:,:,j] + D*c[j]
        elif mode == 'max':
            pass
        elif mode == 'weight':
            pass
        else:
            raise ValueError('mode has to be one of layer, blend or add.')


    if mode == 'add':
        im = np.where(im > 255, 255, im)
    

    bg_mask = np.logical_not(np.any(np.stack([v for v in pages_dict.values()], axis=0), axis=0))
    for j in [0,1,2]:
        im[:,:,j] = np.where(bg_mask, color_dict[bg][j], im[:,:,j])

    
    # im = (im/np.max(im)*255).astype('uint8')
    im = im.astype('uint8')

    Im = Image.fromarray(im)
    if addlegend:
        Im = add_legend(Im, pcolor_dict)
    return Im

def add_legend(im, pcolor_dict, bg=(0,0,0)):
    pagenames = list(pcolor_dict.keys())
    width, height = im.size
    '''
    Add colored legend of element names below an image.
    '''
    # select font
    # if 'darwin' in sys.platform:
    #     Fontttf = "Lao Sangam MN.ttf"
    # else:
    #     Fontttf = "arial.ttf"
    Fontttf = "C059-Roman.otf"
    
    padding = int(height*0.05)
    fontheight = int(padding*0.7)
    font = ImageFont.truetype(Fontttf, fontheight)

    fulltext = "    ".join(pagenames)
    textwidth = font.getsize(fulltext)[0]
    if textwidth > width:
        padding = int(np.floor(width/textwidth*padding))
        fontheight = int(padding*0.7)
        font = ImageFont.truetype(Fontttf, fontheight)

    new_height = height + padding
    Im_new = Image.new(im.mode, (width, new_height), bg)
    Im_new.paste(im, (0,0))
    draw = ImageDraw.Draw(Im_new)
        
    for i, p in enumerate(pagenames):
        c = pcolor_dict[p]
        if c in colors_default.keys():
            c = colors_default[c]
        x = int(width/len(pagenames)*i + width/len(pagenames)/2 - font.getsize(p)[0]/2)
        draw.text((x, height), p,font=font, fill=c)

    return Im_new

def legendbar(c_dict = colors_default, width = 1):
    n = len(c_dict)
    ims = {}
    for i,k in enumerate(c_dict):
        im = np.zeros((n*width, width))
        im[i*width:(i+1)*width, :] = 1
        ims[k] = im
    im = CombineColor(ims, c_dict, bg = 'w', addlegend=False)
    plt.imshow(im)
    
    plt.tick_params(axis='both', which='both', top=False,
                    bottom=False, left=False, right=False,
                    labelbottom=False, labelleft=False, labelright=True)

    plt.yticks(ticks = width*np.arange(n) + np.floor(0.35*width), labels=list(c_dict.keys()))


def getcolornames( includeAbbr = True):
    if includeAbbr:
        return [c for c in colors_default.keys()]
    else:
        return [c for c in colors_default.keys() if len(c)>1]


def axisoff():
    plt.tick_params(axis='both', which='both', top=False,
                    bottom=False, left=False, right=False,
                    labelbottom=False, labelleft=False)



def whiteframe(COLOR = 'white'):
    for spine in plt.gca().spines.values():
        spine.set_edgecolor(COLOR)


from datetime import datetime
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm import tqdm
import io

def MyFun():
    print(':)')
    return '?'

def nonzero(im):
    r = np.ravel(im)
    return r[r>0]

def report(str):
    current_time = datetime.now().strftime("%H:%M:%S")
    print(current_time, "  ", str)

def axisoff():
    plt.tick_params(axis='both', which='both', top=False,
                    bottom=False, left=False, right=False,
                    labelbottom=False, labelleft=False)

def newfig(figsize=[12,12], facecolor=[1,1,1]):
    plt.figure(figsize=figsize, facecolor=facecolor)



def M2E(M, nonzero=True):
    '''
    Convert an adjacency matrix to an edge list.
    '''
    df = pd.DataFrame(M)
    df = df.stack().reset_index()
    if nonzero:
        return df.loc[df.iloc[:,2]>0, :].values
    else:
        return df.values

def E2M(E, v = None, padding=0, dtype=None, return_base = False):
    if v is None and E.shape[1] == 3:
        v = E[:, 2]
    if dtype is None:
        if v is None:
            dtype=bool
        else:
            dtype = v.dtype
    x = E[:, 0].astype(int)
    y = E[:, 1].astype(int)
    xmin = x.min() - padding
    ymin = y.min() - padding
    x_shifted = x - xmin
    y_shifted = y - ymin
    xmax = x_shifted.max() + padding + 1
    ymax = y_shifted.max() + padding + 1
    if v is None:
        nn = np.zeros((xmax, ymax), dtype=bool)
        nn[x_shifted, y_shifted] = 1
    else:
        nn = np.zeros((xmax, ymax), dtype=dtype)
        nn[x_shifted, y_shifted] = v
    if return_base:
        return nn, np.array([xmin, ymin])
    else:
        return nn
        
    

def maskcrop(im, mask=None, padding=None, return_range=False):
    mask = mask.copy()
    im_size = im.shape
    if type(mask)==type(None):
        mask = (im>0)
    mask = mask*1
    
    xleft = np.argwhere(np.any(mask, axis=0))[0][0]
    xright = np.argwhere(np.any(mask, axis=0))[-1][0]

    ytop = np.argwhere(np.any(mask, axis=1))[0][0]
    ybot = np.argwhere(np.any(mask, axis=1))[-1][0]

    if padding == None:
        padding = int(0.05*max(xright-xleft, ybot-ytop))
    xleft = int(max(xleft - padding, 0))
    xright = int(min(xright + padding, im_size[1]))
    ytop = int(max(ytop - padding, 0))
    ybot = int(min(ybot + padding, im_size[0]))
    if len(im.shape) == 3:
        im_new = im[ytop:ybot,xleft:xright,:]
    else:
        im_new = im[ytop:ybot,xleft:xright]
    if return_range:
        return im_new, [ytop,ybot,xleft,xright]
    else:
        return im_new

def imshow(im, cmap='gray'):
    if len(im.shape)==2:
        plt.imshow(im, cmap=cmap)
    else:
        plt.imshow(im)
    plt.axis('off')
    
def splitmask(masks, cell_df, target_col = 'Class'):
    masks_new = {c:np.zeros_like(masks) for c in cell_df[target_col].unique()}

    for cell_id in tqdm(cell_df.Cell):
        c = cell_df.loc[cell_df.Cell == cell_id, target_col].values[0]
        masks_new[c] = np.logical_or(masks_new[c], masks==cell_id)
    return masks_new


def readchannel(data, fn, m, p=99, k='im'):
    ''''Read a channel and cap at p percentile'''
    im = data[fn][k][m]
    im = np.where(im > np.percentile(nonzero(im) , p), np.percentile(nonzero(im) , p), im)
    return im


def uint8(im):
    return (im/max(0.001, np.max(im))*255).astype('uint8')

def dist(a,b):
    return np.sqrt(np.sum(np.square(b-a)))

import seaborn as sns
def paired_palette():
    cp = sns.color_palette('Paired')
    colornames = [a+b  for b in ['b','g','r','o','p','y']for a in ['b','d']]
    cd = {colornames[i]: tuple((np.array(cp[i])*255).astype(int)) for i in range(len(cp))}
    for a in ['b','g','r','o','p','y']:
        cd['m'+a] = tuple(((np.array(cd['b'+a]) + np.array(cd['d' + a]))*0.5).astype(int))
    return cd


import matplotlib.gridspec as gridspec
def hist2d(x=None, y=None, df=None, c1='V1', c2='V2', t1=None, t2=None, bins=128, includezero=True,cmap='inferno', title='histgram', p=75, xmin=None, xmax=None, ymin=None, ymax=None, figsize=(7,7), t1_color = (1,0,0), t2_color = (1, 0,0)):
    '''
    Use either x and y, or df with m1 and m2.
    '''
    plt.figure(facecolor=(1,1,1), figsize = figsize)
    gs = gridspec.GridSpec(5, 5) # ( 10, 12)
    if df is None:
        if x is None or y is None:
            raise SyntaxError('Invalid input data.')
        df = pd.DataFrame({c1:x, c2:y})
    if not includezero:
        df = df.loc[np.logical_and(df[c1]>0, df[c2]>0), :]
    d1 = df[c1]
    d2 = df[c2]
    
    ax0 = plt.subplot(gs[1:, 0:4]) # [6:10, 5:9]
    cc,_,_ = np.histogram2d(d1, d2, bins=bins)
    p = np.percentile(nonzero(cc), p)
    plt.hist2d(d1, d2, bins = bins, cmap = cmap, vmax=p)
    plt.xlabel(c1)
    plt.ylabel(c2)
    plt.xlim([xmin, xmax])
    plt.ylim([ymin, ymax])

    axx = plt.subplot(gs[:1, 0:4]) # [5:6, 5:9]
    plt.hist(d1, bins=bins)
    plt.xlim(ax0.get_xlim())
    plt.tick_params(
        axis='both',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom=False,  
        left=False,
        top=False,        
        right=False,
        labelbottom=False,
        labelleft=False) 
    if t1 is not None:
        ax0.axvline(t1, color=t1_color)
        axx.axvline(t1, color=t1_color)

    axy = plt.subplot(gs[1:, 4:]) # [6:10, 9:10]
    _ = plt.hist(d2, bins=bins, orientation="horizontal")
    plt.ylim(ax0.get_ylim())
    plt.tick_params(
        axis='both',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom=False,  
        left=False,
        top=False,        
        right=False,
        labelbottom=False,        
        labelleft=False) 
    if t2 is not None:
        ax0.axhline(t2, color=t2_color)
        axy.axhline(t2, color=t2_color)

    axx.set_title(title)
    return

def Im2Bytes(Im):
    img_byte_arr = io.BytesIO()
    Im.save(img_byte_arr, format='PNG')
    img_byte_arr = img_byte_arr.getvalue()
    return img_byte_arr